# ParseFDF
# by Curtis Ballard, HP (curtis.ballard@hp.com)
# first release 1 March 2008
# second release 1 December 2010
#    - Renamed from ParseFDF_to_csv to ParseFDF as capability of 
#      original ParseFDF was merged in
#    - Optional text only output for T10/T11 voting or csv output 
#    - Handle strikeout/textedit as a single change instead of a 
#      delete and an insert
#    - Changed from single location field to rectangle for
#      describing location of comment in original PDF file
#    - Handle callouts with arrows and report where the text and
#      arrow pointer should be as separate coordinates
#    - Report responses, accept, reject, etc up to 10 levels deep
#      inside the original comment instead of as new comments
#    - Perl help/man style options using Pod 
#    - Simplistic UTF-8 handling: display code for non-ASCII chars
#      Better long term solution at least for csv would be use 
#      "upgrade" function to convert to all UTF-8
#
# third release Dec, 2011
#    - Search file for comments that have been Accepted/Rejected/Completed using Acrobat
#    - Option to generate new fdf file with comments numbered
#
# Minor update August, 2012
#    - Handle comment numbers with text between the name and number
#      like #HPQ-Late-09/30/12-1 when exporting to csv
#    - Corrected issue where identical comments on multiple pages could
#      cause a name collision which resulted in multiple comments
#      reporting the same comment number in text and csv output
#
# based in part on: ParseFDF for Acrobat 5/6/7 .fdf files
# by Rob Elliott, HP (elliott@hp.com)
# 
# Which was based in part on Ralph Weber's fdf2cmt_pgm.pl
#
# inputs an .fdf file, outputs a .csv file  or .txt file numbering and 
# summarizing the comments optionally outputs a new numbered fdf file
# 
# Usage, in Acrobat (or Acrobat Reader 7/8/9 for some .PDF files),
# Export comments to any <filename>.FDF file
# perl parsefdf.pl [-name XYZ] [-text] <filename>.fdf
#
# The optional -name argument specifies the company/commenter name (e.g., -name HPQ)
# If not specified, it's left blank
#
# The optional -text argument specifies that the output should be raw text
# instead of csv.  If not specified the output will be csv.
#
# <filename>.csv or <filename>.txt contains the output
# 
# Note that the page numbers used are 1..n based; they don't honor
# any alternate page numbering that has been selected
#
# Acrobat 7 removed a lot of useful whitespace, making parsing
# harder.
use Text::Wrap;    # import wrap ()
use Getopt::Long; # import GetOptions()
use Pod::Usage;

my $man = 0;
my $help = 0;
my $name = "";
my @auth_names;
my $textonly = 0;
my $num_fdf = 0;
my $renum_fdf = 0;
my $cntwarn = 0;
my $name_options = 0;
my $name_cnt = "";
GetOptions('help|?' => \$help, man => \$man, 'name:s' => \$name, 
	'text' => \$textonly, 'num' => \$num_fdf, 'numnew' => \$renum_fdf,
        'warn' => \$cntwarn, 'numauth=s' => \@auth_names ) or pod2usage(2);

pod2usage(1) if $help;
pod2usage(-exitstatus => 0, -verbose => 2) if $man;

if (scalar(@auth_names) >= 1) { $name_options++; }
if ($num_fdf == 1) { $name_options++; }
if ($renum_fdf == 1) { $name_options++; }

if ($name_options > 1) {
    print "\nMore than one naming option is set.  These options can't be set together.\n";
    print "    -num option numbers everything\n";
    print "    -numnew numbers comments not already numbers as late comments\n";
    print "    -numauth numbers comments from the specified author(s)\n";
    exit(1);
}


$commentprefix= "";    # prepend before each comment

$infile = $outfile = $outfile2 = shift (@ARGV);

if ($textonly == 1) {
    $Text::Wrap::columns = 74;    # wrap output lines
    # use CRLF for newline when in textmode so simple DOS viewers can handle the output
    $Text::Wrap::separator = "\x0D\x0A> ";
    $outfile =~ s/.fdf/.txt/i;
    # Set append for comments differently for text and csv 
    $append = "\x0D\x0A> ";
} else {
    # Use a narrower text wrap
    $Text::Wrap::columns = 40;    # wrap output lines
    $outfile =~ s/.fdf/.csv/i;
    # Set append for comments differently for text and csv 
    $append = "\x0A";
}

if( length( $infile ) == 0 ) { die "No FDF file to process"; };

print "Processing \"$infile\" into \"$outfile\"\n";

open( IN,  "<$infile" ) || die ("Error opening $infile\n");
binmode IN;    # some .fdfs contain paperclips which include binary streams
open( OUT, ">$outfile" ) || die ("Error opening $outfile\n");
# Have to convert to binary mode for csv to prevent auto LF to CRLF conversion on DOS/Windows
binmode OUT;


# Open a file to create the new fdf if numbering comments
if ($name_options == 1) {
    $outfile2 =~ s/.fdf/-num.fdf/i;
    open( OUTFDF, ">$outfile2" ) || die ("Error opening $outfile2\n");
    print "Writing numbered fdf file as $outfile2\n";
    # need binmode for this one too
    binmode OUTFDF;
}

# Create a hash which will hold the csv file lines to allow searching for matching responses

my %comments_hash = ();
my @comments_list = ();
my $cprefix = "";

# Some special handling to get into csv files compatible with Excel
# Excel allows line breaks in a cell inside a csv file if the line is broken
# with a LF but will start a new line when a line is broken with a CRLF.
# The Perl IO layer will autoconvert a LF to a CRLF if running on a Windows
# system and using textmode so we had to switch to binmode and will have to 
# do special handling for line breaks.

# Lines that need to be ended with CRLF will have it manually entered since 
# we have set the file to binmode so \n wouldn't get autoconverted

# FDF files use CR as line separator
# CR is chr(13) and LF is chr(10)
$normal_crlf = $/;
$/ = chr(13);

# Generate the comment count prefix
if (($commentprefix ne "") || (${name} ne "")) {
    #Only put in this field if something will be there
    $cprefix = "${commentprefix}${name}";
} else {
    $cprefix = "";
}

# Do the real work - put comments into a hash
ParseComments();

# Now print out all the comments from the hash
# Start with a header
if ($textonly != 1) {
    # Module for outputing to csv

    print OUT "Comments on $infile,,,,\x0D\x0A";
    print OUT "State,";
    if ($cprefix ne "") {
        #Only put in this field if something will be there
        print OUT "Company,";
        $cprefix .= ",";
    }

    print OUT "Comment Number,Page,Subtype,Subject,Author,Comment\x0D\x0A";

    # Now iterate through the hash of comments and build the comment lines
    foreach (@comments_list) {
        # Pull the fields out of the hash
        my $next_empty = $comments_hash{$_}[0];
        my $cstate_p = $comments_hash{$_}[1];
        my $number_p = $comments_hash{$_}[2];
        my $page_p = $comments_hash{$_}[3];
        my $subtype_p = $comments_hash{$_}[4];
        my $subj_p = $comments_hash{$_}[5];
        my $author_p = $comments_hash{$_}[6];
        my $comment_p = $comments_hash{$_}[7];

        # For output to csv multi-line comments are in quotes so " must 
        # be changed to ""
        $comment_p =~ s/"/""/g;

        # Now that quotes are escaped put a single quote at the beginning and end
        $comment_p = "\x22" . $comment_p . "\x22";

        # Make the csv line with the first comment - remember that cprefix is a field that may 
        # not contain anything so the comma is built into the variable and it is printed w/o 
        # the comma so I use {} to make certain it parses correctly.
        print OUT "$cstate_p,${cprefix}${number_p},$page_p,$subtype_p,$subj_p,$author_p,$comment_p";

        # Now check for replies
        my $next_pointer = 8;
        while ( $next_pointer < $next_empty ) {
            # A reply is in the next array location, pull it out
            my $c_response = $comments_hash{$_}[$next_pointer++];
            my $c_subtype = $comments_hash{$_}[$next_pointer++];
            my $c_author = $comments_hash{$_}[$next_pointer++];

            # need same handling of quotations as regular comments

            # For output to csv multi-line comments are in quotes so " must 
            # be changed to ""
            $c_response =~ s/"/""/g;
    
            # Now that quotes are escaped put a single quote at the beginning and end
            $c_response = "\x22" . $c_response . "\x22";

            print OUT ",Response $c_subtype by $c_author,$c_response";

        }

        # Manual CRLF to end line for all types of files at the end of a comment
        print OUT "\x0D\x0A"; #CRLF
    }
} else {
    # Text only output, format for easy reading

    print OUT "Comments on $infile\x0D\x0A";
    print OUT "\x0D\x0A---\x0D\x0A";
    if ($cprefix ne "") {
        #Only put in this field if something will be there
        $cprefix .= "-";
    }

    # Now iterate through the hash of comments and build the comment lines
    foreach (@comments_list) {
        # Pull the fields out of the hash
        my $next_empty = $comments_hash{$_}[0];
        my $cstate_p = $comments_hash{$_}[1];
        my $number_p = $comments_hash{$_}[2];
        my $page_p = $comments_hash{$_}[3];
        my $subtype_p = $comments_hash{$_}[4];
        my $subj_p = $comments_hash{$_}[5];
        my $author_p = $comments_hash{$_}[6];
        my $comment_p = $comments_hash{$_}[7];

        print OUT "Comment=${cprefix}${number_p} : State=$cstate_p\x0D\x0A";
        print OUT "Page=$page_p : Comment Type=$subtype_p,$subj_p : Comment Author=$author_p\x0D\x0A> ";
        print OUT "$comment_p";

        # Now check for replies
        my $next_pointer = 8;
	# Set up the response counter to 1
        my $response_no = $next_pointer - 7;
        while ( $next_pointer < $next_empty ) {
            # A reply is in the next array location
            my $c_response = $comments_hash{$_}[$next_pointer++];
            my $c_subtype = $comments_hash{$_}[$next_pointer++];
            my $c_author = $comments_hash{$_}[$next_pointer++];
            print OUT "\x0D\x0A"; #CRLF
            print OUT "==== Response #$response_no, $c_subtype by $c_author ====";
            print OUT "\x0D\x0A"; #CRLF
            print OUT "$c_response";
	    $response_no++;
        }
        # Manual CRLF to end line for all types of files at the end of a comment
        print OUT "\x0D\x0A"; #CRLF

        # Print a blank line and dividing line
        print OUT "\x0D\x0A---\x0D\x0A";

    }
}


# Close FDF/PDF file & restore line separator
close( IN ) || die ("Error closing $infile\n");
close( OUT ) || die ("Error closing $outfile\n");
$/ = $normal_crlf;


# search for author, subtype, etc. in the passed string
# and update global variables if found
sub SearchForStuff {
    my ($line) = shift (@_);

    if ($line =~ /endobj/ ) {
        $endobj=1;
    }

    if ($line =~ /\/Parent/ ) {
        $skipthisobj = 1;
    }

    if ($line =~ /\/T\s*\(([^)]*)\)/ ) {
        $author = $1;
    }

    # check for a PDF resolution statement
    if ($line =~ /\/Contents\s*(\([^)]* set by [^)]*\))/ ) {
        if ($1 =~ /Accepted/) {
            $cstate = "A";
        }
        if ($1 =~ /Rejected/) {
            $cstate = "R";
        }
        if ($1 =~ /Completed/) {
            $cstate = "C";
        }
        if ($1 =~ /Cancelled/) {
            $cstate = "X";
        }
        if ($1 =~ /None/) {
            $cstate = "N";
        }
    }

    if ($line =~ /\/Page\s+(\d+)/ ) {
        $page = $1 + 1; # in Acrobat 7, page numbers are 0-based in fdf
    }

    if ($line =~ /\/Subtype\s*\/(\w+)/ ) {
        $subtype = $1;
    }

    if ($line =~ /\/Subj\s*\(([^)]*)\)/ ) {
        $subj = $1;
    }
        
    # Find out if this is the special case of a text edit where two entries make up one change
    if ($line =~ /\/StrikeOutTextEdit/ ) {
        $textedit = 1;
    }

    # It is possible to put an annotation outside the printable area and get a negative
    # number in the Rectangle so need to handle that in the match below.
    if ($line =~ /\/Rect\s*\[(-*\d*\.\d*)\s*(-*\d*\.\d*)\s*(-*\d*\.\d*)\s*(-*\d*\.\d*)/ ) {
        # Measurements are from lower left corner in points with 72 points to the inch
        # Convert to top left corner and inches

        # Get the raw numbers and save in case we want to report or modify them
        $over_pts = $1;
        $down_pts = $2;
        $over2_pts = $3;
        $down2_pts = $4;

        $over_raw = $over_pts / 72;
        $down_raw = 11 - ($down_pts / 72);
        $over2_raw = $over2_pts / 72;
        $down2_raw = 11 - ($down2_pts / 72);

        # Don't want that much precision so convert to 2 places
        $over = sprintf("%.2f", $over_raw);
        $down = sprintf("%.2f", $down_raw);
        $over2 = sprintf("%.2f", $over2_raw);
        $down2 = sprintf("%.2f", $down2_raw);

        if ($textedit == 1) {
            # Special Case - if this is a "StrikeOutTextEdit" the rectangle defines the space where the next
            # next text should go and the rectangle in the next text just defines the insert caret
            # save the rectangle in another variable for later use
            $over_textedit = $over;
            $down_textedit = $down;
            $over2_textedit = $over2;
            $down2_textedit = $down2;
        }
    }

    # Check for an internal rectangle difference indicating where the text would be for notes etc.
    if ($line =~ /\/RD\s*\[(-*\d*\.\d*)\s*(-*\d*\.\d*)\s*(-*\d*\.\d*)\s*(-*\d*\.\d*)/ ) {
        # These numbers contain adjustments to the Rect numbers to indicate where the
        # text should go
        $text_over_pts = $1;
        $text_down_pts = $2;
        $text_over2_pts = $3;
        $text_down2_pts = $4;
    }

    # Check for a coordinate of where the tip of a pointer should be
    if ($line =~ /\/CL\s*\[(-*\d*\.\d*)\s*(-*\d*\.\d*)/ ) {
        # The CL is 2 or 3 pairs of numbers for point, knee, end or point, end
        # we only care about the point so I didn't pull knee/end values
        $pointer_x = $1;
        $pointer_y = $2;
    }

    # Check for an annotation name (NM) - need for each initial entry
    if ($line =~ /\/NM\s*\(([^)]*)\)/ ) {
        # The NM entry is a text string for the entry name which may be
        # used to tie another entry IRT (In Response To)
        $entry_name = $1;

	# Note that NM is usually unique but is only guaranteed unique
	# on a page so append the page number
        if ($line =~ /\/Page\s+(\d+)/ ) {
           $entry_name .= "-$1"
        }
	   
    }

    # Check for an In Response To (IRT)
    if ($line =~ /\/IRT\s*\(([^)]*)\)/ ) {
        # The IRT entry indicates this is in reponse to the entry with the name
        # Not used for StrikeOutTextEdit
        if ( $textedit != 1) {
                $in_response_to = $1;
        }
	
	# Because NM is only unique on a page and we added the page number,
	# need to do the same for IRT so we can find the original comment
        if ($line =~ /\/Page\s+(\d+)/ ) {
           $in_response_to .= "-$1"
        }
	   
    }
} # SearchForStuff

# perform cleanups on the comment text
sub CleanupLine {
    my $line = shift (@_);

    $line =~ s/\r//g;    # 
    $line =~ s/\n//g;    #
    $line =~ s/\\t/ /g;    # tabs
    $line =~ s/\\$//;    # trailing \

    $line =~ s/\\\)/@/g;    # change \) to a single @ to avoid confusion
    # if the comment contains a @, it will be corrupted
            
    # simplify fancy versions of special characters 
    $line =~ s/\x84/-/g;    # dash  (octal 204)
    $line =~ s/\x85/-/g;    # dash  (octal 205)
    $line =~ s/\x8D/"/g;    # double quote (octal 215)
    $line =~ s/\x8E/"/g;    # double quote (octal 216)
    $line =~ s/\x90/\'/g;    # single quote (octal 90)
    $line =~ s/\xD7/x/g;    # multiply (octal 327)

    # simplify fancy versions of quotes and other special characters that are
    # escaped in Postscript/PDF
    $line =~ s/\\203//g;    # this appears next to 215/216 (quotes)
    $line =~ s/\\204/-/g;    # long dash
    $line =~ s/\\215/\"/g;    # opening double quote
    $line =~ s/\\216/\"/g;    # closing double quote
    $line =~ s/\\220/\'/g;    # single quote
    $line =~ s/\\327/x/g;    # multiply

    return $line;
} # CleanupLine

sub UTFtoText {
    my $print_warn = shift (@_);
    my $this_text = shift (@_);

    # Crude routine to convert a UTF-8 string to ASCII
    # Non-ASCII characters are replaced with \u####

    # Convert string to array of characters
    my @chars = unpack "(a)*", $this_text;

    # Now walk the string - every other char is 0x00 for characters
    # that copy straight over to ASCII, otherwise turn into \u####
    my $newstring = "";
    if ($print_warn == 1) {
        # If this is converting UTF for output into a comment explain how special chars show up
        $newstring = "UTF-8 text \\u#### used for non-ASCII chars\x0A";
    }
    my $shortstring = "";

    for (my $i = 0; $i < @chars;$i += 2) {
        my $first_char = $chars[$i];
        my $second_char = $chars[$i+1];
        if ( $first_char =~ /\x00/ ) {
            # simple case of plain ASCII
            $newstring .= $second_char;
        } else {
            # Not plain ASCII
            $shortstring = sprintf("\\u%2x%2x", ord($first_char), ord($second_char));
            $newstring .= $shortstring;
        }
    }
    return $newstring;
}

sub TexttoUTF {
    my $this_text = shift (@_);

    # Crude routine to convert a ASCII string to UTF-8

    # Convert string to array of characters
    my @chars = unpack "(a)*", $this_text;

    # Now walk the string - insert \x00 between every character
    my $newstring = "";

    for (my $i = 0; $i < @chars; $i++) {
        $newstring .= "\x00";
        $newstring .= $chars[$i];
    }
    return $newstring;
}

# walk through the file
sub ParseComments {
    my ($line);

    # Skip to beginning of a comment line
    while( $line = <IN> ) {
        # if making a numbered fdf file send this line out unmodified, nothing interesting
        if ($name_options == 1) {
            print OUTFDF $line;
        }
        chomp $line;
        if( $line  =~ s/Type\s*/Annot/ ) { 
            last; 
	}
    };

    # Process the comments
    $author = $page = $subtype = $subj = $comment = $comment_line = "";
    $down = $over = $down2 = $over2 = "";
    $over_pts = $over2_pts = $down_pts = $down2_pts = 0;
    $text_over_pts = $text_over2_pts = $text_down_pts = $text_down2_pts = 0;
    $cstate = "O";
    $textedit = 0;
    $number = $this_reviewer_number = 1;
    $non_num_count = 1;
    $new_num = $outnum = $this_num = $uni_num = "";
    while( $line = <IN> ) {
	# Save a copy of the line exactly the way it was in the original fdf
	$savedline = $line;
        chomp $line;

        # parse the simple attributes that precede the Contents attribute
        SearchForStuff ($line);

        # the Contents attribute can be split over multiple lines
        if( $line =~ s/\/Contents\s*\((.*)// ) {
            $line = $1;  # garbage stripped off the left but not the right
            $line = CleanupLine ($line);
            $line =~ s/\).*/\)/;  # strip off anything after any remaining )
            $done = $line =~ /\)/;
            $comment = $line;    # start of the full string

            if ($name_options == 1) {
                # If any of the name options are selected only one is allowed to be selected
                # and we need to keep the contents information so that comment numbers can
                # be inserted.  Don't insert until ready to output so we can figure out
                # who the author is.
                $this_com_fdf .= $savedline;
            }

            # pull in lines until we find a bare )
            while (!$done) {
                unless ($line = <IN>) { last; }
                if ($name_options == 1) {
                    # same as above, keep this line
                    $this_com_fdf .= $line;
                }
                $line = CleanupLine ($line);
                $line =~ s/\s*\)(.*)/\)/;  # strip off anything after any remaining )
                $restofline = $1;    # in Acrobat 6, leftover contains more attributes to parse
                SearchForStuff ($restofline);
                $line =~ s/\s*$//;    # trailing spaces
                $done = $line =~ /\)/;
                $comment .= $line;    # continue
            } # While !done

            $comment =~ s/\)$//;       # get rid of the final )
            $comment =~ s/\\\\/\\/g;   # allow \
            $comment =~ s/\\r/\n/g;    # allow forced CRs
            $comment =~ s/\\n/\n/g;    # allow forced CRs
            $comment =~ s/\\\(/(/g;    # allow (
            $comment =~ s/@/)/g;       # put back the internal )s 

        } else { # if Contents 
	    # if not a Contents line just capture the line as is
            if ($name_options == 1) {
                # same as above, keep this line
                $this_com_fdf .= $savedline;
            }
	} # if Contents

        $comment =~ s/\n+$//;    # get rid of trailing blank lines (allowed in the middle)

        # Handle unicode comments
        if ( $comment =~ /^\xFE\xFF/ ) {
            # Found the Adobe UTF-8 string indicator
            # Now we need to remove the UTF-8 indicator at the start
            $comment =~ s/^\xFE\xFF//;
            $comment = UTFtoText(1, $comment);
        } # if $comment

        if ($endobj && $author && !$skipthisobj ) {
            if ($textedit == 1) { # step 1 means we just found a StrikeOutTextEdit

                # Special case of a StrikeOutTextEdit - shouldn't print out the strike out since
                # it is empty, just increment and print out the rectangle with the next entry.
                #
                # Couldn't use the $skipthisobj case as we need the rectangle from the strikeout to
                # go into the comment from the textedit so we have to pass through here twice.
                #
                # The rectangle for the textedit describes where the caret for inserting the text
                # goes but we want to know where the old text is being struck from which is the 
                # rectangle with the strikeout.

                # Increment to step 2 which tells the code later that the comment follows a
                # StrikeOutTextEdit so we know to use a rectangle already calculated
                $textedit = 2;
            } else {
                # Initialize to an empty comment - we will put some information about the comment
                # into this line and then append the comment itself
                $comment_line = "";

                # Print the rectangle information - because it is from lower left to 
                # upper right and we think upper left to lower right the order printed 
                # is different from pulled

                if ($textedit == 0) {
                    # If not a StrikeOutTextEdit then print the rectangle information
                    if ( $in_response_to eq "") {
                        $comment_line .= "${new_num}In Rectangle (over,down) $over,$down2 to $over2,${down}${append}";
                    } # if $in_response_to
                } elsif ($textedit == 2) {
                    # Special case of an entry following a StrikeOutTextEdit
                    # use the rectangle from the previous entry
                    $comment_line .= "${new_num}In Rectangle (over,down) ";
                    $comment_line .= "$over_textedit,$down2_textedit to $over2_textedit,${down_textedit}${append}";
                    # Now clear textedit as we used it
                    $textedit = 0;
                } else {
                    print STDERR "Got to somewhere we shouldn't have in generating the csv entry.\n"
                } # if $textedit
    
                # Check for special handling of callouts
                if ( $pointer_x > 0 ) { # is there a valid CL (callout line) field?  All parms must be > 0 to be valid.
                    # does this callout have a valid RD (text location) field?  All parms must be > 0 to be valid.
                    if ( $text_over_pts > 0 ) { 
                        $text_over_raw = ($over_pts - $text_over_pts) / 72;
                        $text_down_raw = 11 - (($down_pts - $text_down_pts) / 72);
                        $text_over2_raw = ($over2_pts - $text_over2_pts) / 72;
                        $text_down2_raw = 11 - (($down2_pts - $text_down2_pts) / 72);
    
                        # Don't want that much precision so convert to 2 places
                        $text_over = sprintf("%.2f", $text_over_raw);
                        $text_down = sprintf("%.2f", $text_down_raw);
                        $text_over2 = sprintf("%.2f", $text_over2_raw);
                        $text_down2 = sprintf("%.2f", $text_down2_raw);
                        $comment_line .= "Text at (over,down) $text_over,$text_down2 to $text_over2,${text_down}${append}";
                    } # if $text_over_pts
    
                    # Now find out where this callout is pointing
                
                    $pointer_x_raw = $pointer_x / 72;
                    $pointer_y_raw = 11 - ($pointer_y / 72);
    
                    # Don't want that much precision so convert to 2 places
                    $pointer_x_txt = sprintf("%.2f", $pointer_x_raw);
                    $pointer_y_txt = sprintf("%.2f", $pointer_y_raw);
                    $comment_line .= "Pointing to (over,down) $pointer_x_txt,${pointer_y_txt}${append}";
                } # if $pointer_x
                # If this is a type strikeout we need to copy that information into the comment
                if ($subtype eq "StrikeOut") {
                    $comment_line .= "StrikeOut:$append";
                } # if $subtype
                $comment_line .= wrap "", "", "$comment";
    
                # Check to see if this comment is in response to another in which case we 
                # need to append to an existing comment and will need to go look it up.
                if ( $in_response_to ne "" ) {
                    # Since this is an in response to, just dump the save fdf record if numbering
                    # don't need to put a number in this entry
                    if ($name_options == 1) {
		        print OUTFDF $this_com_fdf;
                        $this_com_fdf = "";
	            }
                    $comment_address = $in_response_to;
    
                    # This is a response so we'll try to pull the pointer to next array field 
                    # from the last comment to see if there is a comment with this name.  If
                    $old_comment = $comments_hash{$comment_address}[0];
    
                    # If no comment at that address may have to follow links back to the original comment so use a while/do loop
                    $stop = 0;
                    $loop = 0;
                    while ((!$old_comment) & ($stop != 1)) {
                        $comment_address = $response_hash{$in_response_to};
    
                        # If we didn't find a previous_irt then stop, replies get put in the fdf twice and we
                        # only use one of them so some irt values don't resolve
                        if (!$comment_address) { $stop = 1; }
    
                        # check for a comment at this address - if we get a pointer to the next empty location
                        # then this hash location contains a valid comment, else we get back undefined
                        $old_comment = $comments_hash{$comment_address}[0];
    
                        # The comment was in response to an "in response to" - we have to take the address of the
                        # response we just found and go find out what it responded to if there wasn't a comment
                        # at this address.
                        if (!$old_comment) {
                            $in_response_to = $comment_address;
                        }
    
                        # Only go back 15 - that seems too many changes and we don't want an infinite loop
                        $loop++;
                        if ($loop == 15) { $stop = 1; }
                    } # while !$old_comment
    
                    # At this point if we have a value in $old_comment then it is an array index
                    # into the array at hash location $comment_address for where to insert the next comment
                    if ( $old_comment ) {
                        $comments_hash{$comment_address}[$old_comment++] = $comment_line;
                        $comments_hash{$comment_address}[$old_comment++] = $subtype;
                        $comments_hash{$comment_address}[$old_comment++] = $author;
    
                        # Put a reply to into the array so adjust the pointer for the next empty location
                        $comments_hash{$comment_address}[0] = $old_comment;
                    } # if $old_comment
                    if ($cstate ne "O") {
                        # The comment state is initialized to "O" if it is changed then we 
                        # found a new state.  The "O" is only output on new comments.
                        $comments_hash{$comment_address}[1] = $cstate;
                    } # if $cstate 
    
                    # Keep a hash of responses and what they responded to as it is possible
                    # to get a response to a response
                    $response_hash{$entry_name} = $in_response_to;
    
                } else {
                    # Not a response - check numbering and then create a new hash entry
                    if ( $in_response_to eq "") {
		        # If numbering and don't have a 'Contents' section create one
		        if (($comment eq "") && ($name_options == 1)) {
			    # No comment - probably no 'Contents' section but be safe
                            if($this_com_fdf =~ /\/C\[.*?\]\/CreationDate/) {
				$this_com_fdf =~ s/(\/C\[.*?\])(\/CreationDate)/$1\/Contents(Empty Comment)$2/;
			    }
		        }
			# search for comment starting with # then anything followed by a '-' with a minimal qualifier
			# to cause stopping at the first '-', using "*?-".  After that search for anything followed by a number.
			# That allows comments in the middle as long as the comment starts with a # and ends with a number
			# such as #HPQ-Late-3
                        if ($comment =~ /^(^#.*?-.*\d*)\D/) {
                            # ASCII Numbered and number is now in $1
                            $this_num = $1;
                        } elsif ($comment =~ /^UTF-8 text.*non-ASCII chars\x0A(#.*?-.*\d*)\D/) {
		            # comment was unicode - had to skip warning that the conversion routine inserted
		            $this_num = $1;
		        } else {
                            # Not numbered - check for renumbering, warning, or doing nothing
                            if ($num_fdf == 1) {
		                # Number everything
		                $outnum = "#${cprefix}-${number}";
                                $new_num = "${outnum}${append}";
                                $this_num = $outnum;
	                    } elsif ($renum_fdf == 1) {
		                # Number new only
		                #WARN - STILL NEED TO FIX DATE ON LATE
		                $outnum = "#Late-${number}";
                                $new_num = "${outnum}${append}";
                                $this_num = $outnum;
	                    } elsif (scalar(@auth_names) >= 1) {
                                # Number only if reviewer name matches
				my $found_auth = 0;
				foreach (@auth_names) {
                                    if ($author eq $_) {
		                        $outnum = "#${cprefix}-${this_reviewer_number}";
                                        $new_num = "${outnum}${append}";
                                        $this_num = $outnum;
                                        $this_reviewer_number++;
					$found_auth = 1;
                                    }
			        }
				if ( $found_auth == 0 ) {
				    # didn't find a matching author
                                    $this_num = $non_num_count++;
			        }
                            } elsif ($cntwarn == 1) {
                                print "CAUTION: Unumbered comment found.  Comment starts with:\n";
                                print "$comment\n";
                                $this_num = $non_num_count++;
                            } else {
                                # doing nothing special, just count
                                $this_num = $non_num_count++;
                            }
	                }
                    }
        
	            # If numbering or renumbering need to insert number and output into new FDF
                    if ($name_options == 1) {
	                # Renumbering something, may be a new number
	                if ($outnum ne "") {
                            # A comment number needs put into the output
		            if ( $comment =~ /^\xFE\xFF/ ) {
                                # Comment is UTF-8
                                $outnum = TexttoUTF($outnum);
                                $this_com_fdf =~ s/(\/Contents\s*\Q(\E\xFE\xFF)/$1$outnum\x00\\r/;
                            } else {
                                # Comment is plain ASCII
                                $this_com_fdf =~ s/(\/Contents\s*\Q(\E)/$1$outnum\\r/;
	                    }
                        }
		        # If numbering always output the line
		        print OUTFDF $this_com_fdf;
                        $this_com_fdf = "";
	            }

                    $comments_hash{$entry_name}[0] = 8;
                    $comments_hash{$entry_name}[1] = $cstate;
                    $comments_hash{$entry_name}[2] = $this_num;
                    $comments_hash{$entry_name}[3] = $page;
                    $comments_hash{$entry_name}[4] = $subtype;
                    $comments_hash{$entry_name}[5] = $subj;
                    $comments_hash{$entry_name}[6] = $author;
                    $comments_hash{$entry_name}[7] = $comment_line;
    
                    # For every new comment line save a lookup value in an array 
                    # so we can pull back for output in order, use zero based array
                    $comments_list[($number-1)] = $entry_name;
    
                    # Increment the total comment number counter
                    $number++;
                } # if $in_response_to
            } # $textedit == 1
        } # if ($endobj && $author && !$skipthisobj )
    
        # reset for next object
        if ($endobj) {
            # Check to see if a cached line for a new fdf file needs output
            if ($this_com_fdf ne "") {
                # A line is in the buffer, output it
		print OUTFDF $this_com_fdf;
                $this_com_fdf = "";
            }

            $author = "";
            $page = "";
            $cstate = "O";
            $subtype = "";
            $subj = "";
            $comment = "";
            $new_num = $outnum = $this_num = $uni_num = "";
            $endobj = 0;
            $skipthisobj = 0;
            $text_over_pts = 0;
            $text_down_pts = 0;
            $text_over2_pts = 0;
            $text_down2_pts = 0;
            $pointer_x = 0;
            $pointer_y = 0;
            $in_response_to = "";
        }
    }; # while line

    # All done - always a few closing lines that will be in the buffer though and need output

    # Check to see if a cached line for a new fdf file needs output
    if ($this_com_fdf ne "") {
        # A line is in the buffer, output it
	print OUTFDF $this_com_fdf;
        $this_com_fdf = "";
    }

} # ParseComments


__END__

=head1 NAME

parsefdf - Parsing utility to convert fdf comments files to text or spreadsheet

=head1 SYNOPSIS

parsefdf [-name -text -help -man] fdf_file

=head1 OPTIONS

=over 8

=item B<-name>    <Company (e.g., HPQ)>

Include the provided company designator with every comment.

=item B<-num>

Number the comments in the fdf file using the company designator and a count.
Output will go to a new fdf file with the same name and followed by -num.

=item B<-numnew>

Number the new comments in the fdf file using today's date and a count.
Output will go to a new fdf file with the same name and followed by -num.

=item B<-numauth> <Author>

Number the comments in the fdf file if the comment was made by <Author>.
The author name must match the author shown in the fdf file exactly.
Multiple author names can be specified on the same command line.
Output will go to a new fdf file with the same name and followed by -num.

=item B<-text>

Output in raw text format suitable for a T10 or T11 letter ballot instead of csv.

=item B<-help>

Print a brief help message and exits.

=item B<-warn>

Prints a caution statement if any comments are found that don't start with '#'.

=item B<-man>

Prints the manual page and exits.

=back

=head1 DESCRIPTION

B<This program> will read the Adobe Acrobat comments from the provided fdf file
and output those comments as text or in csv format for use in a spreadsheet.

=cut
